"""Service layer for registering datasets and models."""
from __future__ import annotations

import datetime
import hashlib
import json
import os
import shutil
from decimal import Decimal
from pathlib import Path

from typing import Iterable
from dataclasses import dataclass,field

from loguru import logger
from sqlalchemy import select, func, distinct
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import aliased

from NepTrainKit import utils

from .database import Database
from .models import (

    Event,
Tag,ModelVersionTag,
    Project,
    ModelVersion,
    StorageRef,
)
from NepTrainKit.core.io.utils import read_nep_in, read_nep_out_file, get_rmse, get_xyz_nframe


@dataclass
class ProjectItem:
    project_id: int
    name:str
    parent_id:int
    model_num:int
    notes:str
    children:list[ProjectItem]=field( default_factory=list)


@dataclass
class TagItem:
    name:str
    tag_id:int
    color:str
    notes:str

@dataclass
class ModelItem:
    model_id: int
    model_type:str
    name:str
    data_size:int
    energy:float
    force:float
    virial:virial
    parent_id:int
    project_id:int
    model_path:str

    notes:str
    created_at:datetime.datetime
    tags:list[TagItem]=field( default_factory=list)

    children:list[ModelItem]=field( default_factory=list)


@dataclass
class TagService:
    """Service class for tag CRUD operations."""
    db: Database

    def create_tag(self, name: str, color: str = "#D1E9FF", notes: str = "") -> TagItem | None:
        """Create a new tag. Returns TagItem or None if name exists."""
        n = name.strip()
        if not n:
            return None
        with self.db.session() as session:
            tag = Tag(name=n, color=color, notes=notes)
            session.add(tag)
            try:
                session.commit()
            except IntegrityError:
                session.rollback()
                return None
            return TagItem(name=tag.name, tag_id=tag.id, color=tag.color, notes=tag.notes)

    def update_tag(self, tag_id: int, **kwargs) -> None:
        """Update tag fields by id."""
        if not kwargs:
            return
        with self.db.session() as session:
            session.query(Tag).filter_by(id=tag_id).update(kwargs)
            session.commit()

    def remove_tag(self, tag_id: int) -> None:
        """Remove tag by id."""
        with self.db.session() as session:
            session.query(Tag).filter_by(id=tag_id).delete()
            session.commit()

    def get_tags(self, **kwargs) -> list[TagItem]:
        """List tags matching filters."""
        with self.db.session() as session:
            tags = (
                session.query(Tag)
                .filter_by(**kwargs)
                .order_by(Tag.created_at)
                .all()
            )
            return [TagItem(name=t.name, tag_id=t.id, color=t.color, notes=t.notes) for t in tags]

def query_set_to_dict(func):
    def wrapper(*args, **kwargs):
        objs = func(*args, **kwargs)
        if not isinstance(objs, (list, tuple)):
            objs = [objs]
        result=[]
        for obj in objs:
            obj_dict = {}
            for column in obj.__table__.columns.keys():
                val = getattr(obj, column)
                if isinstance(val, Decimal):
                    val = float(val)
                if isinstance(val, datetime.datetime):
                    val = val.strftime("%Y/%m/%d %H:%M:%S")
                elif isinstance(val, datetime.date):
                    val = val.strftime("%Y/%m/%d")

                obj_dict[column] = val
            result.append(obj_dict)

        return result
    return wrapper

def _hash_file(path: str) -> str:
    """Return the SHA256 hash of a file."""
    h = hashlib.sha256()
    with open(path, "rb") as f:  # pragma: no cover - simple
        for chunk in iter(lambda: f.read(8192), b""):
            h.update(chunk)
    return h.hexdigest()
def _stat_path(p: Path) -> tuple[int | None, datetime.datetime | None, bool]:
    try:
        st = p.stat()
        return st.st_size, datetime.datetime.fromtimestamp(st.st_mtime), True
    except FileNotFoundError:
        return None, None, False

def _create_storage(session, path: str, scheme: str = "file") -> StorageRef:
    p = Path(path).expanduser().resolve()
    size, mtime, ok = _stat_path(p)

    if scheme == "file":
        uri =  str(p)
        content_hash = utils.sha256_file(p)
    else:
        # cas 策略
        content_hash = utils.sha256_file(p)
        #TODO: 如果拷贝文件 放在那里
        cas_dir = settings.cas_root.expanduser().resolve()
        cas_dir.mkdir(parents=True, exist_ok=True)
        dst = cas_dir / content_hash
        if not dst.exists():
            shutil.copy2(p, dst)
        uri = content_hash
    storage = StorageRef(
        scheme=f"{scheme}://",
        uri=uri,
        size=size,
        content_hash=content_hash,
        mtime=mtime,
        exist_flag=True
    )
    session.add(storage)
    session.flush()
    return storage
@dataclass
class ProjectService:
    """
    项目的service
    """
    db: Database

    def _build_tree(self, node: Project) -> ProjectItem:
        """递归地把 ORM 对象转成 ProjectItem"""
        item = ProjectItem(
            project_id=node.id,
            name=node.name,
            notes=node.notes,
            parent_id=node.parent_id,
            model_num=node.model_num  # 前面 column_property
        )
        # 关键：继续递归

        for child in node.children:
            child_item=self._build_tree(child)
            item.model_num+=child_item.model_num
            item.children.append(child_item)
        return item

    # @query_set_to_dict
    def search_projects(self,**kwargs) -> list[ProjectItem]:
        with self.db.session() as session:
            projects= (session.query(Project)
                        .filter_by(**kwargs)
                        .order_by(Project.created_at)
                        .all())


            return [self._build_tree(p) for p in projects]


    def get_project_by_id(self, project_id: int) -> ProjectItem|None:
        projects = self.search_projects(id=project_id)
        if len(projects) == 0:
            return None
        return projects[0]

    def create_project(self,  name: str, notes: str = "", parent_id: int = None   ) -> Project|None:
        try:
            with self.db.session() as session:
                family = Project(name=name, notes=notes,parent_id=parent_id )
                session.add(family)
                session.commit()
                return family
        except Exception as e:
            logger.error(e)
            return None
    def modify_project(self, project_id: int, name: str, notes: str = "", parent_id: int = None ) -> Project|None:


        with self.db.session() as session:

            session.query(Project).filter_by(id=project_id).update(
                {"name":name,"notes":notes,"parent_id":parent_id}
            )

            session.commit()


    def remove_project(self, project_id: int) -> Project|None:
        with self.db.session() as session:
            session.query(Project).filter_by(id=project_id).delete()


            session.commit()


@dataclass
class ModelService:
    db: Database

    def _project_ids_with_descendants_subq(self, session, root_ids: list[int]):
        """
        返回一个可用于 IN (...) 的子查询，包含 root_ids 以及它们的全部子项目 id。
        支持传入单个或多个 root id。
        """
        P = aliased(Project)

        # 支持多根：先用一个临时表 roots 装入起点
        roots = select(Project.id).where(Project.id.in_(root_ids)).cte("roots")

        # 递归：从根出发向下找 children
        desc = select(roots.c.id.label("id")).cte("desc", recursive=True)
        desc = desc.union_all(
            select(P.id).where(P.parent_id == desc.c.id)
        )
        # 用 select(desc.c.id) 作为 IN 子查询即可
        return select(desc.c.id)
    def search_models_advanced(
            self,
            *,
            project_id: int | list[int] | None = None,
            include_descendants: bool = True,
            parent_id: int | None = None,
            name_contains: str | None = None,
            notes_contains: str | None = None,
            model_type: str | None = None,
            tags_all: Iterable[str] | None = None,
            tags_any: Iterable[str] | None = None,
            tags_none: Iterable[str] | None = None,
            order_by_created_asc: bool = True,
            limit: int | None = None,
            offset: int | None = None,
    ) -> list[ModelItem]:
        names_all = self._normalize_names(tags_all) if tags_all else []
        names_any = self._normalize_names(tags_any) if tags_any else []
        names_none = self._normalize_names(tags_none) if tags_none else []

        with self.db.session() as session:
            mv = ModelVersion
            tg = Tag
            mvt = ModelVersionTag

            q = session.query(mv)

            # ------------ 关键：项目 + 子项目 ------------
            if project_id is not None:
                if isinstance(project_id, int):
                    roots = [project_id]
                else:
                    roots = list(project_id)
                if include_descendants:
                    subq = self._project_ids_with_descendants_subq(session, roots)
                    q = q.filter(mv.project_id.in_(subq))
                else:
                    q = q.filter(mv.project_id.in_(roots))

            # 其他常规过滤
            if parent_id is not None:
                if parent_id<0:
                    q = q.filter(mv.parent_id == None)

                else:

                    q = q.filter(mv.parent_id == parent_id)
            if model_type:
                q = q.filter(mv.model_type == model_type)
            if name_contains:
                q = q.filter(mv.name.like(f"%{name_contains}%"))
            if notes_contains:
                q = q.filter(mv.notes.like(f"%{notes_contains}%"))

            # 标签 OR（任一）
            if names_any:
                q = q.join(mv.tags).filter(tg.name.in_(names_any)).distinct()

            # 标签 AND（全部）
            if names_all:
                q = (q.join(mv.tags)
                     .filter(tg.name.in_(names_all))
                     .group_by(mv.id)
                     .having(func.count(distinct(tg.id)) == len(names_all)))

            # 标签 NOT（排除）
            if names_none:
                sub_exists = (
                    session.query(mvt.model_version_id)
                    .join(tg, tg.id == mvt.tag_id)
                    .filter(mvt.model_version_id == mv.id,
                            tg.name.in_(names_none))
                    .exists()
                )
                q = q.filter(~sub_exists)

            q = q.order_by(mv.created_at.asc() if order_by_created_asc else mv.created_at.desc())
            if offset: q = q.offset(offset)
            if limit:  q = q.limit(limit)

            models = q.all()
            return [self._build_tree(m) for m in models]

    # --- 内部工具：归一化输入标签 ---
    @staticmethod
    def _normalize_names(names: Iterable[str]) -> list[str]:
        # 去空白、去空字符串、去重（大小写不敏感，保留首次出现的原样式）
        seen = set()
        result = []
        for n in names or []:
            if not n:
                continue
            s = n.strip()
            if not s:
                continue
            key = s.lower()
            if key in seen:
                continue
            seen.add(key)
            result.append(s)
        return result

    # --- 内部工具：获取或创建标签（大小写不敏感唯一） ---
    def _get_or_create_tags(self, names: Iterable[str]) -> list["Tag"]:
        names = self._normalize_names(names)
        if not names:
            return []
        with self.db.session() as session:

            # 先查已有的（Tag.name 建议使用 sqlite_collate="NOCASE" 唯一约束）
            existing = session.query(Tag).where(Tag.name.in_(names)).all()
            have_lower = {t.name.lower(): t for t in existing}

            # 需要新建的
            to_create = [Tag(name=n) for n in names if n.lower() not in have_lower]
            if to_create:
                session.add_all(to_create)

                session.flush()


        return list(have_lower.values())
    def _get_or_create_tags_simple(self, names: Iterable[str],session) -> list["Tag"]:
        """单机简化版：查已有 -> 插入缺失 -> flush -> 返回全部"""
        names = self._normalize_names(names)
        if not names:
            return []
        # 1) 已有


        existing = session.query(Tag).where(Tag.name.in_(names)).all()

        have_lower = {t.name.lower(): t for t in existing}
        # 2) 缺失则创建
        to_create = [Tag(name=n) for n in names if n.lower() not in have_lower]
        if to_create:
            session.add_all(to_create)
            session.flush()  # 拿到 id
            session.commit()
        return existing + to_create
    def _build_tree(self, node: ModelVersion) -> ModelItem:
        """递归地把 ORM 对象转成 ModelItem"""

        item = ModelItem(
            model_id=node.id,
            name=node.name,
            notes=node.notes,
            parent_id=node.parent_id,
            project_id=node.project_id,
            model_type=node.model_type,
            data_size=node.data_size,
            energy=node.energy,
            force=node.force,
            virial=node.virial,
            model_path=node.model_path,
            # train_params=node.train_params,
            # calc_params=node.calc_params,
            created_at=node.created_at,


        )
        for tag in node.tags:
            item.tags.append(TagItem(name=tag.name, tag_id=tag.id,notes=tag.notes,color=tag.color))
        # 关键：继续递归
        for child in node.children:
            item.children.append(self._build_tree(child))
        return item


    def get_models_by_project_id(self, project_id: int) -> list[ModelItem]:
        return self.search_models_advanced(project_id=project_id,parent_id=-1)


    def add_version_from_path(self,
                              model_type:str,
                              name:str,
                              path: Path,
                              project_id: int,
                              notes: str = "",
                              tags: list[str] | None = None,

                              parent_id: int | None = None,) -> ModelVersion:
        if model_type.upper()=="NEP":
            model_file=path.joinpath("nep.txt")
            data_file=path.joinpath("train.xyz")
            data_size=get_xyz_nframe(data_file)
            train_params=read_nep_in(path.joinpath("nep.in"))
            energy_array=read_nep_out_file(path.joinpath("energy_train.out"))
            energy = get_rmse(energy_array[:,0],energy_array[:,1])*1000
            force_array=read_nep_out_file(path.joinpath("force_train.out"))
            force = get_rmse(force_array[:,:3],force_array[:,3:])*1000
            virial_array=read_nep_out_file(path.joinpath("virial_train.out"))
            virial = get_rmse(virial_array[:,:6],virial_array[:,6:])*1000
            calc_params={}



            return self.add_version(
                project_id=project_id,
                name=name,
                model_type=model_type.upper(),
                # model_file=model_file,
                # data_file=data_file,
                model_path=path,
                data_size=data_size,
                energy=energy,
                force=force,
                virial=virial,
                # train_params=train_params,
                # calc_params=calc_params,
                notes=notes,
                tags=tags,
                parent_id=parent_id
            )


    def add_version(self,
                    project_id: int,
                    name:str,
                    model_type: str,
                    # model_file: Path|str | None = None,
                    # data_file: Path|str | None = None,
                    model_path: Path|str | None = None,
                    data_size: int | None = None,
                    # train_params: dict | None = None,
                    # calc_params: dict | None = None,
                    energy:float|None = None,
                    force:float|None = None,
                    virial:float|None = None,
                    tags:list[str] | None = None,
                    notes: str = "",
                    parent_id: int | None = None,
                    id:int|None = None
                    ) -> ModelVersion:

        with self.db.session() as session:
            tag_objs = self._get_or_create_tags_simple(tags or [],session)

            project = session.get(Project, project_id)
            if project is None:
                raise ValueError(f"Project {project} not found")
            # data_sr =   None
            # model_sr =   None
            # if model_file:
            #     model_sr  = _create_storage(session,model_file )
            # if data_file:
            #     data_sr  = _create_storage(session,data_file)

            # if parent_id is None:
            #     parent_id = project.active_model_version_id
            if isinstance(model_path, Path):
                model_path=model_path.as_posix()
            version = ModelVersion(
                id=id,
                project_id=project_id,
                name=name,
                model_type=model_type,
                model_path=model_path,
                # model_storage_ref_id = (model_sr.id  if model_sr else None),

                # data_storage_ref_id = (data_sr.id if data_sr else None),
                data_size=data_size,
                # train_params=train_params,

                # calc_params=calc_params,
                energy=energy,
                force=force,
                virial=virial if virial<10000 else 0,
                notes=notes,
                parent_id=parent_id,
                tags=set(tag_objs)
            )
            session.add(version)
            session.flush()


            # project.active_model_version_id = version.id
            # event = Event(
            #     entity_type="data_version",
            #     entity_id=version.id,
            #     action="register",
            #     payload_json=json.dumps({"path": path}),
            # )
            # session.add(event)
            session.commit()
            return version
    def modify_model(self,model_id:int,**kwargs):
        with self.db.session() as session:
            mv = session.get(ModelVersion, model_id)  # 取对象
            if not mv:
                raise ValueError(f"ModelVersion {model_id} not found")
            tag_objs = self._get_or_create_tags_simple(kwargs.pop("tags", []) or [],session)

            # 先更新普通字段
            for key, value in kwargs.items():
                setattr(mv, key, value)

            # 单独更新关系字段
            if tag_objs:
                mv.tags = set(tag_objs)
            session.commit()
    def remove_model(self, model_id: int) :
        with self.db.session() as session:
            session.query(ModelVersion).filter_by(id=model_id).delete()
            session.commit()

class LineageService:
    """Query lineage chains for data and model versions."""

    def __init__(self, db: Database):
        self.db = db

    def _trace(self, cls, start_id: int):
        with self.db.session() as session:
            node = session.get(cls, start_id)
            chain = []
            while node is not None:
                chain.append(node)
                if node.parent_id is None:
                    break
                node = session.get(cls, node.parent_id)
            return chain


    def model_lineage(self, model_version_id: int):
        return self._trace(ModelVersion, model_version_id)



from PySide6.QtCore import QAbstractTableModel, QModelIndex, Qt

class IdNameTableModel(QAbstractTableModel):
    """
    高性能两列表模型：ID / Name
    - load_records_fast(records): 一次性灌入 [(id, name), ...] 或 [{'id':..,'name':..}, ...]
    - load_records_stream(query, batch_size): SQLAlchemy 流式插入（持续插行信号）
    - set_fetch_supplier(supplier): 配合 canFetchMore/fetchMore 的懒加载（滚动再拉）
    """

    __slots__ = ("_rows", "_headers", "_fetch_supplier", "_eof")

    def __init__(self, parent=None):
        super().__init__(parent)
        self._rows = []                 # list[tuple(id, name)]
        self._headers = ("ID", "Name")  # 表头
        self._fetch_supplier = None     # 懒加载供应器: () -> list[tuple]|None
        self._eof = True                # 懒加载是否已到结尾

    # ---------- 基础接口 ----------
    def rowCount(self, parent=QModelIndex()) -> int:
        return 0 if parent.isValid() else len(self._rows)

    def columnCount(self, parent=QModelIndex()) -> int:
        return 0 if parent.isValid() else 2

    def setHorizontalHeaderLabels(self, labels):
        self._headers = tuple(labels)

    def headerData(self, section: int, orientation: Qt.Orientation, role: int = Qt.DisplayRole):
        if role != Qt.DisplayRole:
            return None
        if orientation == Qt.Horizontal:
            if 0 <= section < 2:
                return self._headers[section]
            return None
        # 垂直表头可返回行号（按需）
        if orientation == Qt.Vertical:
            return section + 1
        return None

    def data(self, index, role=Qt.DisplayRole):
        if not index.isValid() or role != Qt.DisplayRole:
            return None
        r, c = index.row(), index.column()
        if not (0 <= r < len(self._rows)) or not (0 <= c < 2):
            return None
        return self._rows[r][c]

    def flags(self, index):
        if not index.isValid():
            return Qt.ItemIsEnabled
        return Qt.ItemIsEnabled | Qt.ItemIsSelectable

    # ---------- 可选：排序（在内存里按列排序；更建议 SQL 端排序） ----------
    def sort(self, column: int, order: Qt.SortOrder = Qt.AscendingOrder):
        if column not in (0, 1):
            return
        self.layoutAboutToBeChanged.emit()
        self._rows.sort(key=lambda x: x[column], reverse=(order == Qt.DescendingOrder))
        self.layoutChanged.emit()

    # ---------- 表头设置 ----------
    def set_header(self, headers=("ID", "Name")):
        self._headers = tuple(headers)

    # ---------- 快速清空 ----------
    def clear(self):
        self.beginResetModel()
        self._rows.clear()
        self._fetch_supplier = None
        self._eof = True
        self.endResetModel()

    # ---------- 方式 A：一次性灌入（内存允许时最快） ----------
    def load_records_fast(self, records):
        """
        records: Iterable of (id, name) 或 {'id':..., 'name':...}
        """
        self.beginResetModel()
        rows = []
        # 若是可预知长度，可预分配（Python list 无显式预分配，这里直接收集）
        for r in records:
            if isinstance(r, dict):
                rows.append((r.get("id"), r.get("name")))
            else:
                # 假定 (id, name)
                rows.append((r[0], r[1]))
        self._rows = rows
        self._fetch_supplier = None
        self._eof = True
        self.endResetModel()

    # ---------- 方式 B：SQLAlchemy 流式/分批灌入（省内存，UI 可见持续增长） ----------
    def load_records_stream(self, query, batch_size=5000):
        """
        query: 例如
          q = (session.query(User.id, User.name)
                 .execution_options(stream_results=True)
                 .yield_per(batch_size))
        """
        # 先重置
        self.beginResetModel()
        self._rows = []
        self._fetch_supplier = None
        self._eof = True
        self.endResetModel()

        buf = []
        def flush(buf_):
            if not buf_:
                return
            start = len(self._rows)
            end = start + len(buf_) - 1
            self.beginInsertRows(QModelIndex(), start, end)
            self._rows.extend(buf_)
            self.endInsertRows()
            buf_.clear()

        for row in query:
            if isinstance(row, (tuple, list)):
                rid, rname = row[0], row[1]
            else:
                rid = getattr(row, "id", None)
                rname = getattr(row, "name", None)
            buf.append((rid, rname))
            if len(buf) >= batch_size:
                flush(buf)
        flush(buf)

    # ---------- 方式 C：懒加载（视图滚动到底部再拉数据） ----------
    def set_fetch_supplier(self, supplier_callable):
        """
        supplier_callable: 无参可调用 -> list[(id, name)] 或 None/[]（表示已到底）
        可把 SQLAlchemy 的游标/生成器封装成每次返回 <= batch_size 的列表。
        """
        self._fetch_supplier = supplier_callable
        self._eof = supplier_callable is None

    def canFetchMore(self, parent=QModelIndex()) -> bool:
        if parent.isValid():
            return False
        return not self._eof and (self._fetch_supplier is not None)

    def fetchMore(self, parent=QModelIndex()):
        if parent.isValid() or self._fetch_supplier is None:
            return
        chunk = self._fetch_supplier()  # 期望返回 list[(id, name)] 或 None
        if not chunk:
            self._eof = True
            return
        start = len(self._rows)
        end = start + len(chunk) - 1
        self.beginInsertRows(QModelIndex(), start, end)
        self._rows.extend(chunk)
        self.endInsertRows()

    # ---------- 小工具：读取一行 ----------
    def get_row(self, row: int):
        if 0 <= row < len(self._rows):
            rid, rname = self._rows[row]
            return {"id": rid, "name": rname}
        return None



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# @Time    : 2024/11/21 14:22
# @Author  : 兵
# @email    : 1747193328@qq.com
import contextlib
import os
import traceback
import numpy as np
import numpy.typing as npt
from PySide6.QtCore import QThread, Signal, QObject
from loguru import logger
from multiprocessing import Process, JoinableQueue, Event

from NepTrainKit import utils
from NepTrainKit.config import Config
from NepTrainKit.core import Structure, MessageManager
from NepTrainKit.core.types import NepBackend

try:
    from NepTrainKit.nep_cpu import CpuNep
except ImportError:
    logger.debug("no found NepTrainKit.nep_cpu")

    try:
        from nep_cpu import CpuNep
    except ImportError:
        logger.debug("no found nep_cpu")


        CpuNep=None
try:
    from NepTrainKit.nep_gpu import GpuNep
except ImportError:
    logger.debug("no found NepTrainKit.nep_gpu")
    logger.debug(traceback.format_exc())
    try:
        from nep_gpu import GpuNep
    except ImportError:
        logger.debug(traceback.format_exc())

        logger.debug("no found nep_gpu")
        GpuNep=None
class NepCalculator():

    def __init__(self, model_file="nep.txt",backend:NepBackend=None,batch_size:int=None):
        super().__init__()
        if not isinstance(model_file, str):
            model_file = str(model_file )
        # print(model_file,backend,batch_size)
        self.initialized = False
        if backend is   None:
            self.backend=NepBackend.AUTO
        else:
            self.backend:NepBackend=backend
        if batch_size is None:
            self.batch_size= 1000
        else:

            self.batch_size:int = batch_size
        self.model_file=model_file
        if CpuNep is None and GpuNep is None:
            MessageManager.send_message_box("Failed to import NEP.\n To use the display functionality normally, please prepare the *.out and descriptor.out files.","Error")
            return

        if os.path.exists(model_file):

            self.load_nep()
            # Probe backend viability and fall back to CPU if GPU is not usable

            self.element_list = self.nep3.get_element_list()
            self.type_dict = {e: i for i, e in enumerate(self.element_list)}
            self.initialized=True
        else:
            self.initialized = False

    def cancel(self):
        self.nep3.cancel()

    def load_nep(self):
        if self.backend == NepBackend.AUTO:
            if not self._load_nep_backend(NepBackend.GPU):
                self._load_nep_backend(NepBackend.CPU)
        elif self.backend == NepBackend.GPU:
            if not self._load_nep_backend(NepBackend.GPU):
                MessageManager.send_warning_message("The NEP backend you selected is GPU, but it failed to load on your device; the program has switched to the CPU backend.")
                self._load_nep_backend(NepBackend.CPU)
        else:
            self._load_nep_backend(NepBackend.CPU)
    def _load_nep_backend(self,backend):
        try:
            with open(os.devnull, 'w') as devnull:
                with contextlib.redirect_stdout(devnull), contextlib.redirect_stderr(devnull):
                    if backend==NepBackend.GPU:
                        if GpuNep is not None:
                            try:
                                self.nep3 = GpuNep(self.model_file)

                            except RuntimeError as e:
                                logger.error( e)
                                MessageManager.send_warning_message(str(e))
                                return False
                            self.nep3.set_batch_size(self.batch_size)
                        else:
                            return False
                    else:
                        # backend==NepBackend.CPU
                        if CpuNep is not None:

                            self.nep3 = CpuNep(self.model_file)
                        else:
                            return False
                    # track the active backend
                    self.backend = backend
                    return True
        except:
            logger.debug(traceback.format_exc())
            return False


    def compose_structures(self, structures:list[Structure]):
        group_size = []
        _types = []
        _boxs = []
        _positions = []
        if  not isinstance(structures,(list,np.ndarray)):
            structures = [structures]
        for structure in structures:
            symbols = structure.get_chemical_symbols()
            _type = [self.type_dict[k] for k in symbols]
            _box = structure.cell.transpose(1, 0).reshape(-1).tolist()
            _position = structure.positions.transpose(1, 0).reshape(-1).tolist()
            _types.append(_type)
            _boxs.append(_box)
            _positions.append(_position)
            group_size.append(len(_type))
        return  _types, _boxs, _positions,group_size

    @utils.timeit
    def calculate(self,
                  structures:list[Structure]
                  )->tuple[npt.NDArray[np.float32],npt.NDArray[np.float32],npt.NDArray[np.float32]]:
        if not self.initialized:
            return np.array([]),np.array([]),np.array([])
        _types, _boxs, _positions,group_size = self.compose_structures(structures)
        self.nep3.reset_cancel()
        try:
            potentials, forces, virials = self.nep3.calculate(_types, _boxs, _positions)
        except Exception:
            logger.debug(traceback.format_exc())
            # If GPU runtime fails, switch to CPU and retry once
            if GpuNep is not None and isinstance(self.nep3, GpuNep):
                MessageManager.send_warning_message("GPU calculation failed; switching to CPU backend.")
                if self._load_nep_backend(NepBackend.CPU):
                    try:
                        potentials, forces, virials = self.nep3.calculate(_types, _boxs, _positions)
                    except Exception:
                        logger.debug(traceback.format_exc())
                        return np.array([]),np.array([]),np.array([])
                else:
                    return np.array([]),np.array([]),np.array([])
            else:
                return np.array([]),np.array([]),np.array([])
        split_indices = np.cumsum(group_size)[:-1]
        #
        potentials=np.hstack(potentials)
        split_potential_arrays = np.split(potentials, split_indices)
        potentials_array = np.array(list(map(np.sum, split_potential_arrays)), dtype=np.float32)
        # print(potentials_array)
        # 处理每个force数组：reshape (3, -1) 和 transpose(1, 0)
        reshaped_forces = [np.array(force).reshape(3, -1).T for force in forces]

        forces_array = np.vstack(reshaped_forces,dtype=np.float32)
        # print(forces_array)

        reshaped_virials = np.vstack([np.array(virial).reshape(9, -1).mean(axis=1) for virial in virials],dtype=np.float32)

        # virials_array = reshaped_virials[:,[0,4,8,1,5,6]]

        return potentials_array,forces_array,reshaped_virials


    @utils.timeit
    def calculate_dftd3(self,
                        structures:list[Structure],
                        functional,
                        cutoff,
                        cutoff_cn)->tuple[npt.NDArray[np.float32],npt.NDArray[np.float32],npt.NDArray[np.float32]]:
        if not self.initialized:
            return np.array([]),np.array([]),np.array([])
        _types, _boxs, _positions,group_size = self.compose_structures(structures)
        self.nep3.reset_cancel()

        try:
            potentials, forces, virials = self.nep3.calculate_dftd3(functional,cutoff,cutoff_cn,_types, _boxs, _positions)
        except Exception:
            logger.debug(traceback.format_exc())
            if GpuNep is not None and isinstance(self.nep3, GpuNep):
                MessageManager.send_warning_message("GPU DFT-D3 failed; switching to CPU backend.")
                if self._load_nep_backend(NepBackend.CPU):
                    try:
                        potentials, forces, virials = self.nep3.calculate_dftd3(functional,cutoff,cutoff_cn,_types, _boxs, _positions)
                    except Exception:
                        logger.debug(traceback.format_exc())
                        return np.array([]),np.array([]),np.array([])
                else:
                    return np.array([]),np.array([]),np.array([])
            else:
                return np.array([]),np.array([]),np.array([])
        split_indices = np.cumsum(group_size)[:-1]
        #
        potentials=np.hstack(potentials)
        split_potential_arrays = np.split(potentials, split_indices)
        potentials_array = np.array(list(map(np.sum, split_potential_arrays)), dtype=np.float32)
        # print(potentials_array)
        # 处理每个force数组：reshape (3, -1) 和 transpose(1, 0)
        reshaped_forces = [np.array(force).reshape(3, -1).T for force in forces]

        forces_array = np.vstack(reshaped_forces,dtype=np.float32)
        # print(forces_array)

        reshaped_virials = np.vstack([np.array(virial).reshape(9, -1).mean(axis=1) for virial in virials],dtype=np.float32)

        # virials_array = reshaped_virials[:,[0,4,8,1,5,6]]

        return potentials_array,forces_array,reshaped_virials

    @utils.timeit
    def calculate_with_dftd3(self,
                             structures:list[Structure],
                             functional,
                             cutoff,
                             cutoff_cn)->tuple[npt.NDArray[np.float32],npt.NDArray[np.float32],npt.NDArray[np.float32]]:
        if not self.initialized:
            return np.array([]),np.array([]),np.array([])
        _types, _boxs, _positions,group_size = self.compose_structures(structures)
        self.nep3.reset_cancel()

        try:
            potentials, forces, virials = self.nep3.calculate_with_dftd3( functional,cutoff,cutoff_cn,_types, _boxs, _positions)
        except Exception:
            logger.debug(traceback.format_exc())
            if GpuNep is not None and isinstance(self.nep3, GpuNep):
                MessageManager.send_warning_message("GPU calculation failed; switching to CPU backend.")
                if self._load_nep_backend(NepBackend.CPU):
                    try:
                        potentials, forces, virials = self.nep3.calculate_with_dftd3( functional,cutoff,cutoff_cn,_types, _boxs, _positions)
                    except Exception:
                        logger.debug(traceback.format_exc())
                        return np.array([]),np.array([]),np.array([])
                else:
                    return np.array([]),np.array([]),np.array([])
            else:
                return np.array([]),np.array([]),np.array([])
        split_indices = np.cumsum(group_size)[:-1]
        #
        potentials=np.hstack(potentials)
        split_potential_arrays = np.split(potentials, split_indices)
        potentials_array = np.array(list(map(np.sum, split_potential_arrays)), dtype=np.float32)
        # print(potentials_array)
        # 处理每个force数组：reshape (3, -1) 和 transpose(1, 0)
        reshaped_forces = [np.array(force).reshape(3, -1).T for force in forces]

        forces_array = np.vstack(reshaped_forces,dtype=np.float32)
        # print(forces_array)

        reshaped_virials = np.vstack([np.array(virial).reshape(9, -1).mean(axis=1) for virial in virials],dtype=np.float32)

        # virials_array = reshaped_virials[:,[0,4,8,1,5,6]]

        return potentials_array,forces_array,reshaped_virials


    def get_descriptor(self,structure:Structure)->npt.NDArray[np.float32]:
        """
        获取单个结构的所有原子描述符
        """
        if not self.initialized:
            return np.array([])
        symbols = structure.get_chemical_symbols()
        _type = [self.type_dict[k] for k in symbols]
        _box = structure.cell.transpose(1, 0).reshape(-1).tolist()

        _position = structure.positions.transpose(1, 0).reshape(-1).tolist()
        self.nep3.reset_cancel()

        descriptor = self.nep3.get_descriptor(_type, _box, _position)

        descriptors_per_atom = np.array(descriptor,dtype=np.float32).reshape(-1, len(structure)).T

        return descriptors_per_atom
    @utils.timeit
    def get_structures_descriptor(self,
                                  structures:list[Structure]
                                  )->npt.NDArray[np.float32]:
        """
        获取结构描述符：原子平均
        返回的已经结构的描述符了 无需平均
        """
        if not self.initialized:
            return np.array([])
        _types, _boxs, _positions, group_size = self.compose_structures(structures)
        self.nep3.reset_cancel()

        descriptor = self.nep3.get_structures_descriptor(_types, _boxs, _positions)

        return np.array(descriptor,dtype=np.float32)

    @utils.timeit
    def get_structures_polarizability(self,
                                      structures:list[Structure]
                                      )->npt.NDArray[np.float32]:
        if not self.initialized:
            return np.array([])
        _types, _boxs, _positions, group_size = self.compose_structures(structures)
        self.nep3.reset_cancel()

        polarizability = self.nep3.get_structures_polarizability(_types, _boxs, _positions)

        return np.array(polarizability,dtype=np.float32)

    def get_structures_dipole(self,
                              structures:list[Structure]
                              )->npt.NDArray[np.float32]:
        if not self.initialized:
            return np.array([])
        self.nep3.reset_cancel()

        _types, _boxs, _positions, group_size = self.compose_structures(structures)

        dipole = self.nep3.get_structures_dipole(_types, _boxs, _positions)

        return np.array(dipole,dtype=np.float32)

Nep3Calculator = NepCalculator

def run_nep_calculator(nep_txt, structures, calculator_type, func_kwargs={},cls_kwargs={},queue=None):
    try:

        nep3 = NepCalculator(nep_txt,**cls_kwargs)
        if calculator_type == 'polarizability':
            result = nep3.get_structures_polarizability(structures)
        elif calculator_type == 'descriptor':
            result = nep3.get_structures_descriptor(structures)
        elif calculator_type == 'dipole':
            result = nep3.get_structures_dipole(structures)
        elif  calculator_type == 'calculate_with_dftd3':
            result = nep3.calculate_with_dftd3(structures,**func_kwargs)
        elif  calculator_type == 'calculate_dftd3':
            result = nep3.calculate_dftd3(structures,**func_kwargs)
        else:
            result = nep3.calculate(structures)
        if queue:
            queue.put(result)  # 将结果通过管道发送给主进程
    except Exception as e:
        logger.error(traceback.format_exc())
        result = np.array([])
        if queue:
            queue.put(result)
    if   queue is  None:
        return result

class NEPProcess(QObject):
    result_signal = Signal( )  # 信号，用于传递进程结果
    error_signal = Signal(str )  # 信号，用于传递错误信息

    def __init__(self ):
        super().__init__()

        self.process = None
        self.use_process = False
        self.func_result = None
    def run_nep3_calculator_process(self,nep_txt, structures, calculator_type="calculate",func_kwargs={},cls_kwargs={},wait=False):
        self.func=run_nep_calculator
        self.queue = JoinableQueue()

        self.input_kwargs={
            "nep_txt": nep_txt,
            "calculator_type": calculator_type,
            "structures": structures,
            "func_kwargs":func_kwargs,
            "cls_kwargs":cls_kwargs,

        }
        #好像在dll释放下gil就不会卡了  这里先统一不使用进程了
        # if len(structures) < 2000:
        #     self.use_process=False
        #     self.input_kwargs["queue"] =  None
        # else:
        #     self.input_kwargs["queue"] = self.queue
        #     self.use_process=True
        # self.use_process=False
        self.input_kwargs["queue"] =  None
        self.input_args=()
        self.func_result:tuple
        self.run()

    def run(self):
        try:
            if self.use_process:
                # 创建并启动进程
                self.process = Process(target=self.func, args= self.input_args ,kwargs = self.input_kwargs,daemon=True)
                self.process.start()
                # 获取结果
                self.func_result = self.queue.get()
            else:
                self.func_result=self.func(*self.input_args,**self.input_kwargs)
            if len(self.func_result) !=0:
                self.result_signal.emit( )
            else:
                self.error_signal.emit("No result returned from process")

        except Exception as e:
            logger.error(traceback.format_exc())
            self.error_signal.emit(f"Error: {str(e)}")
        finally:
            self.queue.close()  # 关闭队列
            # self.queue.join()  # 等待队列任务完成
            if self.process is not None :
                self.process.terminate()  # 确保进程被终止

                self.process.join()
                # self.process.close()

    def stop(self):
        """强制终止进程并清理资源"""
        return
        try:

            if self.process is not None:
                if self.process.is_alive():

                    self.process.terminate()

                    self.process.join()
            self.queue.cancel_join_thread()
        except Exception as e:
            logger.error(traceback.format_exc())

if __name__ == '__main__':
    structures = Structure.read_multiple(r"D:\Desktop\nep\nep-data-main\2023_Zhao_PdCuNiP\train.xyz")
    nep = NepCalculator(r"D:\Desktop\nep\nep-data-main\2023_Zhao_PdCuNiP\nep.txt")
    nep.calculate(structures)



#!/usr/bin/env python 
# -*- coding: utf-8 -*-
# Lightweight, lazy exports to avoid heavy imports at startup.
from __future__ import annotations

from typing import Any

__all__ = [
    'MessageManager',
    'Structure', 'process_organic_clusters', 'get_clusters',
    'CardManager', 'load_cards_from_directory',
]


def __getattr__(name: str) -> Any:
    if name == 'MessageManager':
        from .message import MessageManager as _M
        return _M
    if name in ('Structure', 'process_organic_clusters', 'get_clusters'):
        from .structure import Structure as _S, process_organic_clusters as _P, get_clusters as _G
        return {'Structure': _S, 'process_organic_clusters': _P, 'get_clusters': _G}[name]
    if name in ('CardManager', 'load_cards_from_directory'):
        from .card_manager import CardManager as _C, load_cards_from_directory as _L
        return {'CardManager': _C, 'load_cards_from_directory': _L}[name]
    raise AttributeError(name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2025/6/18 13:21
# @Author  : 兵
# @email    : 1747193328@qq.com
from itertools import combinations

import numpy as np
from PySide6.QtWidgets import QFrame, QGridLayout
from qfluentwidgets import BodyLabel, ComboBox, ToolTipFilter, ToolTipPosition, CheckBox, EditableComboBox

from NepTrainKit import utils
from NepTrainKit.core import CardManager, process_organic_clusters, get_clusters
from NepTrainKit.custom_widget import SpinBoxUnitInputFrame
from NepTrainKit.custom_widget.card_widget import MakeDataCard
from scipy.stats.qmc import Sobol

@CardManager.register_card
class PerturbCard(MakeDataCard):
    group = "Perturbation"
    card_name= "Atomic Perturb"
    menu_icon=r":/images/src/images/perturb.svg"
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setTitle("Make Atomic Perturb")
        self.init_ui()

    def init_ui(self):
        self.setObjectName("perturb_card_widget")
        self.engine_label=BodyLabel("Random engine:",self.setting_widget)
        self.engine_type_combo=ComboBox(self.setting_widget)
        self.engine_type_combo.addItem("Sobol")
        self.engine_type_combo.addItem("Uniform")
        self.engine_type_combo.setCurrentIndex(1)

        self.engine_label.setToolTip("Select random engine")
        self.engine_label.installEventFilter(ToolTipFilter(self.engine_label, 300, ToolTipPosition.TOP))

        self.optional_frame = QFrame(self.setting_widget)
        self.optional_frame_layout = QGridLayout(self.optional_frame)
        self.optional_frame_layout.setContentsMargins(0, 0, 0, 0)
        self.optional_frame_layout.setSpacing(2)

        self.optional_label=BodyLabel("Optional",self.setting_widget)
        self.organic_checkbox=CheckBox("Identify organic", self.setting_widget)
        self.organic_checkbox.setChecked(False)
        self.optional_label.setToolTip("Treat organic molecules as rigid units")
        self.optional_label.installEventFilter(ToolTipFilter(self.optional_label, 300, ToolTipPosition.TOP))



        self.optional_frame_layout.addWidget(self.organic_checkbox,0,0,1,1)

        self.scaling_condition_frame = SpinBoxUnitInputFrame(self)
        self.scaling_condition_frame.set_input("Å",1,"float")
        self.scaling_condition_frame.setRange(0,1)
        self.scaling_radio_label=BodyLabel("Max distance:",self.setting_widget)
        self.scaling_condition_frame.set_input_value([0.3])
        self.scaling_radio_label.setToolTip("Maximum displacement distance")
        self.scaling_radio_label.installEventFilter(ToolTipFilter(self.scaling_radio_label, 300, ToolTipPosition.TOP))

        self.num_condition_frame = SpinBoxUnitInputFrame(self)
        self.num_condition_frame.set_input("unit",1,"int")
        self.num_condition_frame.setRange(1,10000)
        self.num_condition_frame.set_input_value([50])

        self.num_label=BodyLabel("Max num:",self.setting_widget)
        self.num_label.setToolTip("Number of structures to generate")

        self.num_label.installEventFilter(ToolTipFilter(self.num_label, 300, ToolTipPosition.TOP))

        self.settingLayout.addWidget(self.engine_label,0, 0,1, 1)
        self.settingLayout.addWidget(self.engine_type_combo,0, 1, 1, 2)

        self.settingLayout.addWidget(self.optional_label, 1, 0, 1, 1)
        self.settingLayout.addWidget(self.optional_frame,1, 1, 1, 2)

        self.settingLayout.addWidget(self.scaling_radio_label, 2, 0, 1, 1)

        self.settingLayout.addWidget(self.scaling_condition_frame, 2, 1, 1,2)

        self.settingLayout.addWidget(self.num_label,3, 0, 1, 1)

        self.settingLayout.addWidget(self.num_condition_frame,3, 1, 1,2)



    def process_structure(self, structure):
        structure_list = []
        engine_type = self.engine_type_combo.currentIndex()
        max_scaling = self.scaling_condition_frame.get_input_value()[0]
        max_num = int(self.num_condition_frame.get_input_value()[0])
        identify_organic = self.organic_checkbox.isChecked()

        n_atoms = len(structure)
        dim = n_atoms * 3

        # 生成扰动因子
        if engine_type == 0:
            sobol_engine = Sobol(d=dim, scramble=True)
            perturbation_factors = (sobol_engine.random(max_num) - 0.5) * 2
        else:
            perturbation_factors = np.random.uniform(-1, 1, (max_num, dim))

        # 识别团簇（如启用）
        if identify_organic:
            clusters, is_organic_list = get_clusters(structure)
            # 预先构建聚类索引（避免每次遍历）
            organic_clusters = [cluster for cluster, is_org in zip(clusters, is_organic_list) if is_org]
            inorganic_clusters = [cluster for cluster, is_org in zip(clusters, is_organic_list) if not is_org]

        orig_positions = structure.positions

        for i in range(max_num):
            delta = perturbation_factors[i].reshape(n_atoms, 3) * max_scaling
            new_positions = orig_positions + delta  # 默认：全部扰动

            if identify_organic:
                # 初始化为原始坐标，随后分组替换
                new_positions = orig_positions.copy()

                # 有机分子：整体平移

                for cluster in organic_clusters:  # pyright:ignore
                    cluster_delta = delta[cluster[0]]
                    new_positions[cluster] += cluster_delta  # 利用 NumPy 索引批量加

                # 无机分子：逐原子扰动（其实和默认行为相同，略冗余，但保留逻辑清晰）
                for cluster in inorganic_clusters:  # pyright:ignore
                    new_positions[cluster] += delta[cluster]

            # 构造新结构
            new_structure = structure.copy()
            new_structure.set_positions(new_positions)
            new_structure.wrap()
            config_str = f" Perturb(distance={max_scaling}, {'uniform' if engine_type == 1 else 'Sobol'})"
            new_structure.info["Config_type"] = new_structure.info.get("Config_type", "") + config_str
            structure_list.append(new_structure)

        return structure_list

    def to_dict(self):
        data_dict = super().to_dict()


        data_dict['engine_type'] = self.engine_type_combo.currentIndex()
        data_dict["organic"]=self.organic_checkbox.isChecked()
        data_dict['scaling_condition'] = self.scaling_condition_frame.get_input_value()

        data_dict['num_condition'] = self.num_condition_frame.get_input_value()
        return data_dict

    def from_dict(self, data_dict):
        super().from_dict(data_dict)

        self.engine_type_combo.setCurrentIndex(data_dict['engine_type'])

        self.scaling_condition_frame.set_input_value(data_dict['scaling_condition'])

        self.num_condition_frame.set_input_value(data_dict['num_condition'])
        self.organic_checkbox.setChecked(data_dict.get("organic", False))
